// Robin Hood Web Server - A web server for BeOS
// Copyright (C) 1999 Simon Huet

// This program is free software; you can redistribute it and/or 
// modify it under the terms of the GNU General Public License 
// as published by the Free Software Foundation; either version 2 
// of the License, or (at your option) any later version. 

// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
// GNU General Public License for more details. 

// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// Contact Info:
// Author: Simon Huet
// E-mail: detourds@bellsouth.net
// Postal Address: 3250 Matilda St, Miami, FL 33133, USA

#include <string.h>
#include <string>
#include <stdio.h>
#include <File.h>
#include <Directory.h>
#include <NodeInfo.h>
#include <stdio.h>
#include <Path.h>
#include <Alert.h>

//Query Module Library //////
#include <ctype.h>		   //	
#include <stdlib.h>		   //
#include <errno.h>		   //
#include <string.h>		   //
#include <dirent.h>		   //
#include <fs_info.h>	   //
#include <fs_query.h>	   //
#include <fs_attr.h>	   //
#include <Entry.h>		   //
#include <TypeConstants.h> //
/////////////////////////////


#include "HTTPMessage.h"
#include "parsedate.h"
#include "ByteRanges.h"
#include "HTTPRealm.h"
#include "RequestPB.h"
#include "RHModuleInterface.h"
#include "HTTPUtils.h"
#include "StringUtils.h"
#include "DataIOUtils.h"

#include "RHQ_error_msg.h"
#include "dummy_messages.h"

#define	GET	'GETM';
#define	POST 'POST';

// Functions
status_t send_file( RequestPB *pb, char * filepath, char *filetype);

// RHQueryData init

// DATA STRUCTS
typedef struct{	char *name; char *value; int  mode;}query_attributes;
typedef struct{	char *name;	char *value;}query_operators;
typedef struct{	char *name;	char *query;}query_def;
typedef struct{	char *part;}body_parts;

typedef struct{
	char 		*name;
	char		*content_type;
	char 		*header;
	char		*body;
	body_parts 	*bodyparts;
	int			bodyparts_count;
	char 		*footer;
	int			record;
	bool		all;
	bool		dummy;
	bool		stream;
}report_def;

// HTTP STRUCT
typedef struct{	char *name;	char *value; }http_value_fields;
typedef struct{ http_value_fields *field; int  count; }http_env_fields;
 

// CLASS
class RHQueryData
{

public:
							RHQueryData(void);
	void 					Query(BMallocIO *body);
	void 					TokenQuery(const char *query_string);
	void					Format_BodyPart(BMallocIO *body, const char* template_body);					
	void 					readfile(void);
	char 					*get_file_stream_path(int report_index);
	
	// ***DATA*** //		
	query_attributes 		*attrs;
	int 					attr_count;
	query_operators 		*ops;
	int 					op_count;
	query_def 				*queries;
	int 					query_count;
	report_def 				*reports;
	int 					report_count;
	char 					*filetype;
	char 					*drive;
	char					*default_query;
	char					*default_report;
	bool					record_report;
	RequestPB 				*pb;
	// ***HTTP*** //
	http_env_fields 		*http_env;
	// ***CONTROL*** //
	bool					has_error;
	int						error_mode;
	bool					has_dummy;
	bool					has_file_to_stream;
	char					*error_message;
	const char 				*error_report;
	bool					output_file;
	int						current_line_number;
	FILE 					*fd;
	
private:

	char 					*Get_QueryString(const char *arg);
	char 					*create_query(void);
	void 					define_attr(char *line, int *line_id);
	void 					define_operator(char *line, int *line_id);
	void 					define_query(char *line, int *line_id);
	void 					read_report(char *name, int *line_id);
	void 					parse_line(char *line, int *line_id);
	void					create_err_msg(const char* error_msg);
};



bool can_handle_resource( const char *mimeType, http_method method, int32 *priority )
{

	if( strcmp( "text/RHQScript", mimeType ) == 0 )
	{
		*priority = 10;
		return true;
	}
	else
		return false;
}


//////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////// QUERY MODULE FUNCTIONS ///////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////////////
// create_uncase_arg
// Format 'arg' to an uncase querey arguments 
// ex: create_uncase_arg("case") will return will be transform to "[Cc][Aa][Ss][Ee]"
//////////////////////////////////////////////////////////////////////////////////////
char *create_uncase_arg(char *arg){
	char *uncase_exp = NULL;
	unsigned i;
	bool mem_allocated = false;
	
	for (i=0; i<strlen(arg); i++){
		if (isalpha(arg[i])){ // if A-Z or a-z
			if (mem_allocated == false){
				uncase_exp = (char *)malloc(6);
				sprintf(uncase_exp, "[%c%c]", tolower(arg[i]), toupper(arg[i]));
				mem_allocated = true;
			}else{
				uncase_exp = (char *)realloc(uncase_exp, strlen(uncase_exp)+6);
				sprintf(uncase_exp, "%s[%c%c]", uncase_exp, tolower(arg[i]), toupper(arg[i]));
			}
		}else{ //else - just copy
			if (mem_allocated == false){
				uncase_exp = (char *)malloc(2);
				sprintf(uncase_exp, "%c", arg[i]);
				mem_allocated = true;
			}else{
				uncase_exp = (char *)realloc(uncase_exp, strlen(uncase_exp)+2);
				sprintf(uncase_exp, "%s%c", uncase_exp, arg[i]);
			}
		}
	}
	return uncase_exp;
}

//////////////////////////////////////////////////////////////////////////////////////
// URLDecode & URLEncode
// decode an URL encoded string or URL encode a string. 
// This is standard functions, nothing special. 
//////////////////////////////////////////////////////////////////////////////////////
char *URLDecode(const char *String){

	char achar;
	string str = "";
	int encharcnt = 0;
	
	for (uint i=0; i < strlen(String); i++){
		achar = 0;
		if (String[i] == '%'){
			//FIRST DIGIT
			if ((String[i+1]>=65) && (String[i+1]<=70)) achar = 16*(String[i+1]-65+10);
			else if ((String[i+1]>=97) && (String[i+1]<=102)) achar = 16*(String[i+1]-97+10);
			else achar = 16*String[i+1];
				
			//SECOND DIGIT
			if ((String[i+2]>=65) && (String[i+2]<=70)) achar += (String[i+2]-65+10);
			else if ((String[i+2]>=97) && (String[i+2]<=102)) achar += (String[i+2]-97+10);
			else achar += atoi(&String[i+2]);
			str += achar;
			encharcnt+=2;
			
		}else if (String[i] == '+')	str += ' ';
		else{
			if (encharcnt > 0)	encharcnt--;
			else	str += String[i];
		}
	}
	char *URLEncoded;
	URLEncoded = (char *)malloc(strlen(str.c_str())+2);
	sprintf(URLEncoded, "%s", str.c_str());
	return URLEncoded;

}
char *URLEncode(const char *String){
	//  URL ENCODING STUFF  //
	
	unsigned char achar;
	int echar;
	char *URLEncoded;


	for (uint i=0; i<strlen(String); i++){
		achar = (unsigned char) String[i];
		echar = (int) achar;
		
		if (((echar>=33)&&(echar<=46)) ||
			((echar>=58)&&(echar<=64)) ||
			((echar>=91)&&(echar<=96)) ||
			((echar>=123)&&(echar<=126)) ){
			if (i==0){
				URLEncoded = (char *)malloc(strlen(String)+4);
				sprintf(URLEncoded, "%%%i", (echar/16));
				if ((echar%16) < 10) sprintf(URLEncoded, "%s%i", URLEncoded, (echar%16));
				else sprintf(URLEncoded, "%s%c", URLEncoded, (echar%16-10+65));			
			}else{
				URLEncoded = (char *)realloc(URLEncoded, strlen(URLEncoded)+4);
				sprintf(URLEncoded, "%s%%%i", URLEncoded, (echar/16));
				if ((echar%16) < 10) sprintf(URLEncoded, "%s%i", URLEncoded, (echar%16));
				else sprintf(URLEncoded, "%s%c", URLEncoded, (echar%16-10+65));
			}
		}else{
			if (i==0){
				URLEncoded = (char *)malloc(strlen(String)+2);
				if (achar == ' ') sprintf(URLEncoded, "+");
				else sprintf(URLEncoded, "%c", achar);
			}else{
				URLEncoded = (char *)realloc(URLEncoded, strlen(URLEncoded)+2);
				if (achar == ' ') sprintf(URLEncoded, "%s+", URLEncoded);
				else sprintf(URLEncoded, "%s%c", URLEncoded, achar);
			}	
		}
	}
	return URLEncoded;
	// END URL ENCODING //
}
char *FormatArgs(const char *args, int mode){
	//mode value:
	//0 - Normal string
	//1 - URLed string
	//2 - HTMLed string (work in progress)
	
	switch (mode){
		case 0:
			return ((char *)args);
			break;	
		case 1:
			return (URLEncode(args));
			break;
	}
	
}


RHQueryData::RHQueryData(void)
{
	attr_count=0;
	op_count=0;
	query_count=0;
	report_count=0;
	has_error = false;
	has_dummy = false;
	error_mode = 0;
}


void
RHQueryData::create_err_msg(const char* error_msg){
	if (has_error == false){
		error_message = (char *)error_msg;
	}
	has_error = true;
	
}

//////////////////////////////////////////////////////////////////////////////////////
// TokenQuery
// Decode the value-paired name-value inside the querystring (?name=value) 
// This is a standard function that may be reuse for something else 
/////////////////////////////////////////////////////////////////////////////////////

void
RHQueryData::TokenQuery(const char *query_string)
{
	char *trk;
	char *tmp = NULL;
	char *query;
	int count;

	//bool stop;
	
	http_env = (http_env_fields *)malloc(sizeof(http_env_fields));
	http_env->count = 0;
	
	query = (char *)malloc(strlen(query_string));
	strcpy(query, URLDecode(query_string));
	
	trk = strtok(query, "&");
	
	while (trk != NULL){
		count = http_env->count;
		if (count == 0) {
			http_env->field = (http_value_fields *)
							malloc(sizeof(http_value_fields));
		}else{
			http_env->field = (http_value_fields *)realloc(
				http_env->field, 
				sizeof(http_value_fields)*(count+1));
		}
		//NAME
		http_env->field[count].name = (char *)malloc(strlen(trk)+2);
		strcpy(http_env->field[count].name, trk);
		trk = strtok(NULL, "&");
		http_env->count++;	
	}
	
	for (int i=0; i<http_env->count; i++) {

		trk = strtok(http_env->field[i].name, "=");
		http_env->field[i].name = trk;

		trk = strtok(NULL, "&");
		
		if ((trk == NULL)) trk = "";
		http_env->field[i].value = trk;

	}
}
//////////////////////////////////////////////////////////////////////////////////////
// Get_QueryString
// Decode the value-paired name-value inside the querystring (?name=value) 
// This is a standard function that may be reuse for something else 
//////////////////////////////////////////////////////////////////////////////////////
char *
RHQueryData::Get_QueryString(const char *arg){
	bool found = false;

	for (uint i=0; i<http_env->count; i++){
		if (!strcmp(http_env->field[i].name, arg)){
			found = true;
			return http_env->field[i].value;
		}
	}

	if (found == false) return NULL;
}
//////////////////////////////////////////////////////////////////////////////////////
// create_query
// create the query string accordindly to each parameters 
// 
//////////////////////////////////////////////////////////////////////////////////////
char *
RHQueryData::create_query(void){
	char *ret, *tmp, *trk;

	bool part=0;
	bool end=false;
	bool found=false;
	char *value;
	int i;
	int query_index = 0;
	
	if (default_query != NULL) {
		for (i=0;i<query_count;i++){
			if (!strcmp(default_query, queries[i].name))
				query_index = i;
		}
	}
	ret=(char *)malloc(4096);
	tmp=(char *)malloc(strlen(queries[query_index].query)+1);
	strcpy(tmp, queries[query_index].query);
	trk = strtok(tmp, " ");
	sprintf(ret, "(");
	do {
		if (trk == NULL){
			end = true;
		}else{
			
			if (part == 0){ //Attribute
				found = false;
				for (i=0;i<attr_count;i++){
					if (!strcmp(attrs[i].name, trk)){
						found = true;
						break;
					}
				}
				//Should do something if found = false
				//If found then add the name to query
				if (found == true){

					if (strcmp(attrs[i].value, ""))	value = attrs[i].value;
					else value = "*";

					sprintf(ret, "%s(%s", ret, attrs[i].name);
					switch (attrs[i].mode){
						case 0:	sprintf(ret, "%s==\"*%s*\")", ret, create_uncase_arg(value)); 	break;
						case 1:	sprintf(ret, "%s==\"%s\")", ret, create_uncase_arg(value)); 	break;
						case 2:	sprintf(ret, "%s!=\"%s\")", ret, create_uncase_arg(value)); 	break;
						case 3:	sprintf(ret, "%s==\"%s*\")", ret, create_uncase_arg(value)); 	break;
						case 4:	sprintf(ret, "%s==\"*%s\")", ret, create_uncase_arg(value)); 	break;
					}
				}
				part = 1;
			} else { // Operator
				found = false;
				for (i=0;i<op_count;i++){
					if (!strcmp(ops[i].name, trk)){
						found = true; break;
					}
				}
				//Should do something if found = false
				//If Found then add operator to query
				if (found == true){
					sprintf(ret, "%s%s", ret, ops[i].value);
				}
				part = 0;
			}
		trk = strtok(NULL, " ");	
		}
	} while (!end);
	sprintf(ret, "%s&&(BEOS:TYPE==\"%s\"))", ret, filetype);
	
	return ret;
}


//////////////////////////////////////////////////////////////////////////////////////
// define_attr
// parse each attribute defined in the input file into memory. 
// 
//////////////////////////////////////////////////////////////////////////////////////
void 
RHQueryData::define_attr(char *line, int *line_id)
{
	//char *ptr;
	char *name;
	char *value;
	char *mode;
	string str_line;
	str_line = line;
	
	name = strtok(line, "{");
	value = strtok(NULL, ",");
	mode = strtok(NULL, "}");

	if (!strncmp(value, "Value", 5)){
		value = strtok (&value[6], ")");
	}else if (!strncmp(value, "Query", 5)){
		value = strtok(&value[6], ")");
		value = Get_QueryString(value);
	}else{

		if (!has_error){
			char *err;
			err = (char *)malloc(512);
			sprintf(err, err010, value, name, current_line_number);
			error_report = RHQ_error_msg;
			create_err_msg(err);
			error_mode = 1;
		}
	}	
	
	
	if (!strncmp(mode, "Mode", 4)) {
		mode = strtok (&mode[5], ")");
	}
	
	if (attr_count == 0) {
		attr_count++;
		attrs = (query_attributes *)malloc(sizeof(query_attributes));
	}else{
		attr_count++;
		attrs = (query_attributes *)realloc(attrs, sizeof(query_attributes)*attr_count);
	}
	if (value != NULL){
		attrs[attr_count-1].name = (char *)malloc(strlen(name)+1);
		strcpy(attrs[attr_count-1].name, name);
		attrs[attr_count-1].value = (char *)malloc(strlen(value)+1);
		strcpy(attrs[attr_count-1].value, value);
		attrs[attr_count-1].mode = atoi((const char *)mode);
	}else{
		if (!has_error){
			name = strtok((char *)str_line.c_str(), "{");
			value = strtok(NULL, ",");
			mode = strtok(NULL, "}");
			str_line = value;
			value = strtok(&value[6], ")");
			char *err;
			err = (char *)malloc(512);
			sprintf(err, err110, str_line.c_str(), value, current_line_number);
			error_report = RHQ_error_msg;
			error_mode = 2;
			create_err_msg(err);
		}
	}
	return;
}


//////////////////////////////////////////////////////////////////////////////////////
// define_operator
// parse each operator defined in the input file into something more pratical to
// deal with. 
//////////////////////////////////////////////////////////////////////////////////////
void 
RHQueryData::define_operator(char *line, int *line_id)
{
	char *name;
	char *value;
	
	name = strtok(line, "{");
	value = strtok(NULL, "}");
	if (!strncmp(value, "Value", 5)) {
		value = strtok (&value[6], ")");
	}

	if (!strncmp(value, "Query", 5)){
		value = Get_QueryString(strtok(&value[6], ")"));
	}
	
	if (!strcmp(value, "AND")) value = "&&";
	else if (!strcmp(value, "OR")) value = "||";
	//else GENERATE ERROR MSG;
	
	if (op_count == 0) {
		op_count++;
		ops = (query_operators *)malloc(sizeof(query_operators));
	}else{
		op_count++;
		ops = (query_operators *)realloc(ops, sizeof(query_operators)*op_count);
	}

	ops[op_count-1].name = (char *)malloc(strlen(name)+1);
	strcpy(ops[op_count-1].name, name);
	ops[op_count-1].value = (char *)malloc(strlen(value)+1);
	strcpy(ops[op_count-1].value, value);	
	
	return;

}

void 
RHQueryData::define_query(char *line, int *line_id)
{
	char *name;
	char *query;
	
	name = strtok(line, "{");
	query = strtok(NULL, "}");
	
	if (query_count == 0) {
		query_count++;
		queries = (query_def *)malloc(sizeof(query_def));
	}else{
		query_count++;
		queries = (query_def *)realloc(queries, sizeof(query_def)*query_count);
	}

	queries[query_count-1].name 	= (char *)malloc(strlen(name)+1);
	strcpy(queries[query_count-1].name, 	name);
	queries[query_count-1].query 	= (char *)malloc(strlen(query)+1);
	strcpy(queries[query_count-1].query, 	query);
	
	return;
}






void 
RHQueryData::readfile(void)
{
  	char *ret;
  	char line[512];
	int fcls;
	unsigned i;
	current_line_number=0;
	
	if (!fd){
		return;
	}else{
  		do {
    		ret = fgets(line, sizeof(line)-1, fd);
    		current_line_number++;
    		line[sizeof(line) - 1] = '\0';
    		line[strlen(line) - 1] = '\0';
    		if ((ret != NULL) && strlen(line)) {
    			//Removing beginning whitespace
				for (i=0; i<strlen(line); i++) {
					if (!isspace(line[i])) break;
    			}
    			//Parse line to analyze
				parse_line(&line[i], &current_line_number);
      		}
    	} while (!feof(fd));
		fcls = fclose(fd);
	}
}



void 
RHQueryData::read_report(char *name, int *line_id)
{

  	char *ret;
  	char line[512];
  	char *Record;
  	char *Stream;
  	char *Content_Type = NULL;
  	bool all = true;
  	bool stream = false;
//	int fcls;
	unsigned i;
	bool end_print = false;
	int mode=0;
	body_parts 	*bodyparts = NULL;
	int body_parts_count=0;
	bool dummy = false;
	
	bool memallocated_head = false;
	bool memallocated_body = false;
	bool memallocated_foot = false;
	
	char *head = NULL;
	char *body = NULL;
	char *foot = NULL;
	
	if (!fd){
		return;
	}else{
  		do {
  			
    		ret = fgets(line, sizeof(line)-1, fd);
    		
    		line[sizeof(line) - 1] = '\0';
    		line[strlen(line) - 1] = '\0';
    		if ((ret != NULL)) {
    			//Removing beginning whitespace
				for (i=0; i<strlen(line); i++) {
					if (!isspace(line[i])) break;
    			}
    			//Detecting end of report
      			if (!strncmp(&line[i], "}", 1)){
      				end_print = true;
   				}else{
   					//Determining if we will print all the record found or just one
   					//FILTER
   					if ( (mode == 0) && (!strncmp(&line[i], "FILTER", 6)) ){
						for (i=i+6; i<strlen(line); i++) {
							if (!isspace(line[i])) break;
		    			}
		    			Record = (char *)malloc(strlen(line)-i+1);
		    			strcpy(Record, &line[i]);

		    			if (!strncmp(Record, "Value", 5)){
		    				Record = strtok(&Record[6], ")");
		    			}else if(!strncmp(Record, "Query", 5)){
		    				Record = Get_QueryString(strtok(&Record[6], ")"));
		    				if (Record == NULL) Record = "ALL";
		    			}
		    			if (!strcmp(Record, "ALL")) all = true;
		    			else all = false;

   					}
   					//STREAM
   					else if ( (mode == 0) && (!strncmp(&line[i], "STREAM", 6)) ){
						for (i=i+6; i<strlen(line); i++) {
							if (!isspace(line[i])) break;
		    			}
		    			Stream = (char *)malloc(strlen(line)-i+1);
		    			strcpy(Stream, &line[i]);

		    			if (!strncmp(Stream, "Value", 5)){
		    				Stream = strtok(&Stream[6], ")");
		    			}else if(!strncmp(Stream, "Query", 5)){
		    				Stream = Get_QueryString(strtok(&Stream[6], ")"));
		    			}
		    			
		    			if (Stream == NULL)	stream = false;
		    			else  				stream = true;

   					}
   					//TYPE
   					else if ( (mode == 0) && (!strncmp(&line[i], "TYPE", 4)) ){
						for (i=i+4; i<strlen(line); i++) {
							if (!isspace(line[i])) break;
		    			}
		    			Content_Type = (char *)realloc(Content_Type, strlen(line)-i+1);
		    			strcpy(Content_Type, &line[i]);

		    			if (!strncmp(Content_Type, "Value", 5)){
		    				Content_Type = strtok(&Content_Type[6], ")");
		    			}else if(!strncmp(Record, "Query", 5)){
		    				Content_Type = Get_QueryString(strtok(&Content_Type[6], ")"));
		    				if (Content_Type == NULL) Content_Type = "text/plain";
		    			}
		    			
   					}

   					else if ( (mode == 0) && (!strncmp(&line[i], "HEAD", 4)) ) mode = 1;
   					else if ( (mode == 1) && (!strncmp(&line[i], "EOF", 3)) ) mode = 2;
   					else if ( (mode == 2) && (!strncmp(&line[i], "[", 1)) ){
   						mode = 21;
   						body_parts_count++;
   						if (body_parts_count == 1) bodyparts = (body_parts *)malloc(sizeof(body_parts));
   						else bodyparts = (body_parts *)realloc(bodyparts, sizeof(body_parts)*body_parts_count);
   						memallocated_body = false;
   					}
   					else if ( (mode ==21) && (!strncmp(&line[i], "]", 1)) ) mode = 2;
   					else if ( (mode == 2) && (!strncmp(&line[i], "FOOT", 4))) mode = 3;
   					else if ( (mode == 3) && (!strncmp(&line[i], "EOF", 3)) ) end_print = true;
   					else if ( (mode == 0) && (!strncmp(&line[i], "DUMMY", 5)) ) mode = 4;
   					else if ( (mode == 4) && (!strncmp(&line[i], "EOF", 3)) ) end_print = true;
   					else{
   						if (mode == 1){
   							if (!memallocated_head){
   								head = (char *)malloc(strlen(&line[i])+2);
   								strcpy(head, &line[i]);
   								memallocated_head = true;
   							}else{ 
   								head = (char *)realloc(head, strlen(head)+strlen(&line[i])+2);
   								strcat(head, "\n");
   								strcat(head, &line[i]); 
   							}
   						}
   						
   						if (mode == 2){
   							if (!memallocated_body){
   								body = (char *)malloc(strlen(&line[i])+2);
   								strcpy(body, &line[i]);
   								memallocated_body = true;
   							}else{ 
   								body = (char *)realloc(body, strlen(body)+strlen(&line[i])+2);
   								strcat(body, "\n");
   								strcat(body, &line[i]); 
   							}
   						}
   						if (mode == 21){

   							if (!memallocated_body){
   								bodyparts[body_parts_count-1].part = (char *)malloc(strlen(&line[i])+2);
   								strcpy(bodyparts[body_parts_count-1].part, &line[i]);
   								memallocated_body = true;
   							}else{
   								bodyparts[body_parts_count-1].part = (char *)realloc(bodyparts[body_parts_count-1].part, 
   									strlen(bodyparts[body_parts_count-1].part) + strlen(&line[i]) + 2);
   								strcat(bodyparts[body_parts_count-1].part, "\n");
   								strcat(bodyparts[body_parts_count-1].part, &line[i]); 
   							}
   							
   						}
   						if (mode == 3) {
   							if (!memallocated_foot){
   								foot = (char *)malloc(strlen(&line[i])+2);
   								strcpy(foot, &line[i]);
   								memallocated_foot = true;
   							}else{ 
   								foot = (char *)realloc(foot, strlen(foot)+strlen(&line[i])+2);
   								strcat(foot, "\n");
   								strcat(foot, &line[i]); 
   							}
   						}
   						if (mode == 4){
   							dummy = true;
   							if (!memallocated_head){
   								head = (char *)malloc(strlen(&line[i])+2);
   								strcpy(head, &line[i]);
   								memallocated_head = true;
   							}else{ 
   								head = (char *)realloc(head, strlen(head)+strlen(&line[i])+2);
   								strcat(head, "\n");
   								strcat(head, &line[i]); 
   							}
   						}
   						end_print = false;
   					}	
   				}
      		}
    	} while ((!feof(fd))&&(!end_print));
    	
		
		
		if (report_count == 0) {
			report_count++;
			reports = (report_def *)malloc(sizeof(report_def));
		}else{
			report_count++;
			reports = (report_def *)realloc(reports, sizeof(report_def)*report_count);
		}
		
		reports[report_count-1].name = name;
		
		if (Content_Type == NULL){
			reports[report_count-1].content_type = "text/plain";
		}else{
			reports[report_count-1].content_type = Content_Type;
		}
		
		if (!stream){
			reports[report_count-1].header = head;
			if (!dummy){
				if (body_parts_count>0){
					reports[report_count-1].bodyparts = bodyparts;
				}else{
					reports[report_count-1].body = body;		
				}
				reports[report_count-1].bodyparts_count = body_parts_count;
				
				reports[report_count-1].footer = foot;
				reports[report_count-1].all = all;
				if (!all) reports[report_count-1].record = atoi(Record);
				else reports[report_count-1].record = 1;
				reports[report_count-1].dummy = false;
			}else{
				reports[report_count-1].dummy = true;
				has_dummy = true;
			}
			reports[report_count-1].stream = false;
		}else{
			if (Stream == NULL){
				reports[report_count-1].record = 1;
				reports[report_count-1].stream = false;
			}else{
				reports[report_count-1].record = atoi(Stream);
				reports[report_count-1].stream = true;
			}
		}

	}

	return;
	
}



void 
RHQueryData::parse_line(char *line, int *line_id)
{
	char 	*word;
	char	*buffer;
	//char	attrline[256];
	
	uint 	i;
	if (line == NULL) return;

	
  	// if passed nothing but whitespace, just return
	for (i=0; i<strlen(line); i++) {
		if (!isspace(line[i])) break;
    }

	if (i>=strlen(line)) return;
	if (!strncmp(&line[i], "#", 1)) return;
	
	//Backup line before manipulating it
	buffer = (char *)malloc(strlen(line)+2);
	strcpy(buffer, line);
	//get the first word
	word = strtok(buffer, " 	");
	buffer = (char *)malloc(strlen(line)+2);
	strcpy(buffer, line);
	
	//parse the filetype and drive value
	if (!strcmp(word, "filetype")){
		word = strtok(NULL, " =	");
		filetype = (char *)malloc(strlen(word)+2);
		strcpy(filetype, word);
	}
	if (!strcmp(word, "drive")){
		word = strtok(NULL, " =	");
		drive = (char *)malloc(strlen(word)+2);
		strcpy(drive, word);
	}
	if (!strcmp(word, "default_query")){
		word = strtok(NULL, " =	");
		if (!strncmp(word, "Query", 5)){
			if (Get_QueryString(strtok(&word[6], ")")) != NULL){
				default_query = (char *)malloc(strlen(Get_QueryString(strtok(&word[6], ")")))+2);
				strcpy(default_query, Get_QueryString(strtok(&word[6], ")")));
			}else default_report = NULL;
		}else if (!strncmp(word, "Value", 5)){
			default_query = (char *)malloc(strlen(word)+2);
			strcpy(default_query, strtok(&word[6], ")"));
		}
	}
	if (!strcmp(word, "default_report")){
		word = strtok(NULL, " =	");
		if (!strncmp(word, "Query", 5)){
			if (Get_QueryString(strtok(&word[6], ")")) != NULL){
				default_report = (char *)malloc(strlen(Get_QueryString(strtok(&word[6], ")")))+2);
				strcpy(default_report, Get_QueryString(strtok(&word[6], ")")));
			}else default_report = NULL;
		}else if (!strncmp(word, "Value", 5)){
			default_report = (char *)malloc(strlen(word)+2);
			strcpy(default_report, strtok(&word[6], ")"));
		}
	}
	//detect define function
	
	if (!strcmp(word, "defattr")){
		word = strtok(NULL, "");
		for (i=0; i<strlen(word); i++) {
			if (!isspace(word[i])) break;
    	}
    	define_attr(&word[i], line_id);
	}
	if (!strcmp(word, "defop")){
		word = strtok(NULL, "");
		for (i=0; i<strlen(word); i++) {
			if (!isspace(word[i])) break;
    	}
    	define_operator(&word[i], line_id);
	}
	if (!strcmp(word, "defquery")){
		word = strtok(NULL, "");
		for (i=0; i<strlen(word); i++) {
			if (!isspace(word[i])) break;
    	}
    	define_query(&word[i], line_id);
	}
	if (!strcmp(word, "defreport")){
		word = strtok(NULL, "{");
		for (i=0; i<strlen(word); i++) {
			if (!isspace(word[i])) break;
    	}
    	record_report = true;
    	read_report(&word[i], line_id);
	}
}

void
RHQueryData::Format_BodyPart(BMallocIO *body, const char* template_body)
{
	
	//TXT flow control variable
	uint curchar_inc=0;
	uint char_inc=0;
	uint char_index=0;
	
	int format_mode=0;
	bool end_loop;
	string arg, body_generated;
	char *tmp=NULL;
	char fieldValue[1024];
	//Cache the size of the body
	uint body_len = strlen(template_body);
	
	body_generated = "";
	
	for (uint i=0; i<body_len-curchar_inc; i++){
		char_index = i + curchar_inc;
		
		// Detect '<@=' opening tag 
		if (!strncmp(&template_body[char_index], "<@=", 3)){

			char_inc = 0; 
			end_loop = false;
			arg = "";
			//Get the argument tag <@=argument>
			while (!end_loop){
				if ((template_body[char_index+3+char_inc]!='>')){
					arg += template_body[char_index+3+char_inc];
					char_inc++;
				}else{
					end_loop = true;
					char_inc++;
				}
			}
			format_mode=0;
			if (!strncmp(arg.c_str(), "URL", 3)){
				arg = strtok((char *)&arg.c_str()[4], "]");
				format_mode=1;
			}
		
			if (!strncmp(arg.c_str(),"Env", 3)){
				if (!strcmp(strtok((char *)&arg.c_str()[4], ")"), "QUERY_STRING"))
					body_generated += FormatArgs(pb->brURI->query, format_mode);
				else if (!strcmp(strtok((char *)&arg.c_str()[4], ")"), "REMOTE_ADDR"))
					body_generated += FormatArgs(pb->request->GetRemoteHost(), format_mode);
				else if (!strcmp(strtok((char *)&arg.c_str()[4], ")"), "SERVER_NAME"))
					body_generated += FormatArgs(pb->brURI->host, format_mode);
				else if (!strcmp(strtok((char *)&arg.c_str()[4], ")"), "SERVER_PORT")){
					sprintf( fieldValue, "%u", pb->request->GetPort() );
					body_generated += FormatArgs(fieldValue, format_mode);
				}else if (!strcmp(strtok((char *)&arg.c_str()[4], ")"), "SERVER_SOFTWARE"))
					body_generated += FormatArgs("RobinHood", format_mode);
			}else if (!strncmp(arg.c_str(), "QueryString", 11)){
				//(new BAlert("","Debug Query String", "OK"))->Go();
				//tmp = (char *)realloc(tmp, strlen(arg.c_str())+2);
				//sprintf(tmp, "%s", arg.c_str());				
				arg = strtok((char *)&arg.c_str()[12], ")");
				tmp = Get_QueryString(arg.c_str());

				if (tmp != NULL)
					body_generated += FormatArgs(tmp, format_mode);
			}else if (!strncmp(arg.c_str(), "Message", 7)){
				if ((has_error) && (error_mode > 2))
					body_generated += FormatArgs(error_message, format_mode);
			}
			
			curchar_inc += char_inc+2;
		}else{
			//NORMAL CHARACTERS-TAGS

			char_inc = 0; 
			end_loop = false;
			arg = "";
			int len = strlen(template_body);
			while ((!end_loop) && ((char_index+char_inc) < len)){
				
				if (!strncmp(&template_body[char_index+char_inc], "<@=", 3)){
					end_loop = true;
				}else{
					arg += template_body[char_index+char_inc];
					char_inc++;
				}
			}
			body_generated += arg;
			curchar_inc += char_inc-1;

		}
	}
	body->Write(body_generated.c_str(), body_generated.length());
}

void 
RHQueryData::Query(BMallocIO *body)
{

	int			body_counts=0;
	int			report_index;
	char 		*predicate;
	bool		end_loop;
	

	//C attribute variables
	attr_info   attrib;	
	char 		filename[PATH_MAX];
	BNode 		*node = NULL;
	dev_t 		d[256];
	memset(d, 0, sizeof(d));
	d[0] = dev_for_path(drive);
	
	//TXT buffer variables
	char *template_body = NULL;
	template_body = (char *)malloc(0);
	template_body[0] = 0;
	
	string arg;
	string body_generated;
	char *tmp = NULL;
	
	//Query variables
	DIR *query;
	dirent *match;
	
	//TXT flow control variable
	uint curchar_inc;
	uint char_inc;
	uint char_index;
	
	//ATTR TYPE
	char 	attr_type_string[B_ATTR_NAME_LENGTH];
	char	attr_type_string_empty[B_ATTR_NAME_LENGTH];
	strcpy(attr_type_string_empty, "");
	int32	attr_type_int32=0;
	int32	attr_type_int32_empty=0;
	
	//QUERY
	bigtime_t 		startTime, elapsedTime;
	startTime = system_time();
	predicate = create_query(); 
	query = fs_open_query(d[0], predicate, 0);
	
	//FORMAT
	int format_mode=0;
	report_index = 0;
	
	if (default_report != NULL) {
		for (int i=0;i<report_count;i++){
			if (!strcmp(default_report, reports[i].name))
				report_index = i;
		}
	}
		
	
	
	if (query){
		
		int file_index=0;
		int file_id;
		pb->Logprintf("@# RHQ Generating Body\n");

		//Count elapsed time in copying body + getting next node
		bigtime_t 		cpybody_startTime, cpybody_elapsedTime;
		cpybody_elapsedTime = system_time() - system_time();
		
		//Count elapsed reading and processing <@= tag
		bigtime_t 		tag_startTime, tag_elapsedTime;
		tag_elapsedTime = system_time() - system_time();
		
		//Count elapsed reading and processing "normal" char
		bigtime_t 		char_startTime, char_elapsedTime;
		char_elapsedTime = system_time() - system_time();
		
		body_generated = "";
		
		while(match = fs_read_query(query)) {
			
			file_index++;
			file_id = file_index -1;
			if ((reports[report_index].all) || 
				((!reports[report_index].all) && (reports[report_index].record == file_id))){
				
				cpybody_startTime = system_time();
				get_path_for_dirent(match, filename, PATH_MAX);
	        	node = new BNode((const char *) filename);
	        	if (node && node->InitCheck() == B_OK) {
	          		node->SetTo((const char *) filename);
	          	}
				if (reports[report_index].bodyparts_count>0){
					if (body_counts == reports[report_index].bodyparts_count) body_counts = 0;
					body_counts++;
					template_body = reports[report_index].bodyparts[body_counts-1].part;
				}else{
					template_body = reports[report_index].body;
				}
				cpybody_elapsedTime += system_time() - cpybody_startTime;
				
				curchar_inc=0;
				for (uint i=0; i<strlen(template_body)-curchar_inc; i++){
					char_index = i + curchar_inc;
					
					// Detect '<@=' opening tag 
					if (!strncmp(&template_body[char_index], "<@=", 3)){
						tag_startTime = system_time();
						char_inc = 0; 
						end_loop = false;
						arg = "";
						//Get the argument tag <@=argument>
						while (!end_loop){
							if ((template_body[char_index+3+char_inc]!='>')){
								arg += template_body[char_index+3+char_inc];
								char_inc++;
							}else{
								end_loop = true;
								char_inc++;
							}
						}
						format_mode=0;
						if (!strncmp(arg.c_str(), "URL", 3)){
							arg = strtok((char *)&arg.c_str()[4], "]");
							format_mode=1;
						}
						
						
						if (!strncmp(arg.c_str(),"Env", 3)){
							if (!strcmp(strtok((char *)&arg.c_str()[4], ")"), "QUERY_STRING"))
								body_generated += FormatArgs(pb->brURI->query, format_mode);
							else if (!strcmp(strtok((char *)&arg.c_str()[4], ")"), "REMOTE_ADDR"))
								body_generated += FormatArgs(pb->request->GetRemoteHost(), format_mode);
						}else if (!strncmp(arg.c_str(), "QueryString", 11)){
							arg = strtok((char *)&arg.c_str()[12], ")");
							tmp = Get_QueryString(arg.c_str());
			
							if (tmp != NULL)
								body_generated += FormatArgs(tmp, format_mode);
							
						}else if (!strcmp(arg.c_str(),"File:Index")){
							tmp = (char *)malloc(32);
							sprintf(tmp, "%i", file_index);
							body_generated += FormatArgs(tmp, format_mode);
						}else if (!strcmp(arg.c_str(),"File:ID")){
							tmp = (char *)malloc(32);
							sprintf(tmp, "%i", file_id);
							body_generated += FormatArgs(tmp, format_mode);
						}else{
							//ATTRIBUTES
							node->GetAttrInfo(arg.c_str(), &attrib);
							//B_STRING_TYPE
		          			if (attrib.type == B_STRING_TYPE){
		          				attr_type_string = attr_type_string_empty;
		          				if ((node->ReadAttr(arg.c_str(), (type_code) 0, (off_t) 0, (void *) &attr_type_string,255))!=0)
		          				body_generated += FormatArgs(attr_type_string, format_mode);
		          			}
		          			//B_INT32_TYPE
		          			if (attrib.type == B_INT32_TYPE){
		          				attr_type_int32 = attr_type_int32_empty;
		          				if ((node->ReadAttr(arg.c_str(), (type_code) 0, (off_t) 0, (void *) &attr_type_int32, 32))!=0)
		          				tmp = (char *)realloc(tmp, 32);
		          				sprintf(tmp, "%i", attr_type_int32);
		          				body_generated += FormatArgs(tmp, format_mode);
		          			}
							//free(arg);
						}
						curchar_inc += char_inc+2;
						tag_elapsedTime += system_time() - tag_startTime;
					}else{
						//NORMAL CHARACTERS-TAGS
						char_startTime = system_time();
						char_inc = 0; 
						end_loop = false;
						arg = "";
						int len = strlen(template_body);
						while ((!end_loop) && ((char_index+char_inc) < len)){
							
							if (!strncmp(&template_body[char_index+char_inc], "<@=", 3)){
								end_loop = true;
							}else{
								arg += template_body[char_index+char_inc];
								char_inc++;
							}
						}
						body_generated += arg;
						curchar_inc += char_inc-1;
						char_elapsedTime += system_time() - char_startTime;
					}
				}
				cpybody_startTime = system_time();
				node->~BNode();
				cpybody_elapsedTime += system_time() - cpybody_startTime;
			}
		}
		
		fs_close_query(query);
		
		if (file_index > 0){
			char_startTime = system_time();
			body->Write(body_generated.c_str(), body_generated.length());
			char_elapsedTime += system_time() - char_startTime;
			elapsedTime = system_time() - startTime;
			
			int32 		milliSeconds = elapsedTime/100;
			int32		record_count = (int32) file_index;
			int32		cpybody_mSeconds = cpybody_elapsedTime/1000;
			int32		tag_mSeconds = tag_elapsedTime/100;
			int32		char_mSeconds = char_elapsedTime/100;
			
			pb->Logprintf( "@# RHQ cpybody elapsed time : %ld ms at %ld 1/10th ms./record\n", 
				cpybody_mSeconds/10, cpybody_mSeconds/record_count);
			pb->Logprintf( "@# RHQ tag elapsed time : %ld ms at %ld 1/10th ms./record\n", 
				tag_mSeconds/10, tag_mSeconds/record_count);
			pb->Logprintf( "@# RHQ char elapsed time : %ld ms at %ld 1/10th ms./record\n", 
				char_mSeconds/10, char_mSeconds/record_count);
			pb->Logprintf( "@# RHQ body done in %ld ms, %ld record(s) at %ld 1/10th ms./record(s)\n", 
				milliSeconds/10, record_count, milliSeconds/record_count);
		}else{
			pb->Logprintf( "@# RHQ No match found" );

			error_message = (char *)no_file_found;
			has_error = true;
			error_mode = 3;

		}
	}
}


char * 
RHQueryData::get_file_stream_path(int report_index)
{
	//C attribute variables
	attr_info   attrib;	
	int 		file_id=0;
	char 		filename[PATH_MAX];
	bool		found = false;
	char 		*predicate;
	dev_t 		d[256];
	memset(d, 0, sizeof(d));
	d[0] = dev_for_path(drive);
	
	//Query variables
	DIR *query;
	dirent *match;
	
	//QUERY
	predicate = create_query(); 
	query = fs_open_query(d[0], predicate, 0);

	if (query){
		while( (!found) && (match = fs_read_query(query)) ) {
			if (reports[report_index].record == file_id){	
				get_path_for_dirent(match, filename, PATH_MAX);
				found = true;
			}else file_id++;
		}
	}
	
	if (found) return (char *)&filename;
	else return NULL;
		
}



status_t handle_request( RequestPB *pb )
{
	
	HTTPResponse	response;
	char			fieldValue[4096];
	char			headBuffer[1024];
	int32			contentLength;	
	BMallocIO		body;
	BMallocIO		error_body;
	
	
	// Temporary buffers
	
	BFile			theFile;//, gzFile
	BPath			absPath;
	//bool			gz = false;
	
	bigtime_t 		RHQstartTime, RHQdeltaTime;
	RHQstartTime = system_time();	
	pb->closeConnection = false;
	absPath.SetTo( pb->webDirectory->Path(), pb->brURI->path );
	
	// Check Authorization
	if( pb->authenticate &&	!pb->realms->Authenticate( pb->request, &response, pb->brURI->path, absPath.Path(), S_IROTH ) )
	{
		pb->Logprintf( "%ld Status-Line: %s\n", pb->sn, response.GetStatusLine() );
		return B_OK;
	}
	
	pb->Logprintf("@# RHQuery version 1.0 starting\n");
	// Was the file found?
	if( 1 )
	{

		RHQueryData *QueryData;
		QueryData = new RHQueryData();
		QueryData->pb = pb;
		QueryData->TokenQuery(pb->brURI->query);
		QueryData->fd = fopen(absPath.Path(),"r");
		QueryData->readfile();
		fclose(QueryData->fd);

		
		// Add Content
		int report_index = 0;
		if (QueryData->has_error == false){	
			if ((QueryData->has_error == false) && (QueryData->default_report != NULL)) {
				for (int i=0;i<QueryData->report_count;i++){
					if (!strcmp(QueryData->default_report, QueryData->reports[i].name))
						report_index = i;
				}
			}
			if (!QueryData->reports[report_index].stream){
				QueryData->Format_BodyPart( &body, QueryData->reports[report_index].header );
				if (!QueryData->reports[report_index].dummy) {
					QueryData->Query(&body);
					QueryData->Format_BodyPart( &body, QueryData->reports[report_index].footer );
				}
			}
		}
		
		if (!QueryData->reports[report_index].stream){
			
			// Add Date header
			time_t			now;
			struct tm 		*brokentime;
			
			now = time( NULL );
			brTimeLock.Lock();
			brokentime = gmtime( &now );
			strftime(fieldValue, 256, kHTTP_DATE, brokentime);
			brTimeLock.Unlock();
			
			response.AddHeader( kHEAD_DATE, fieldValue );		
			
	
			//response.SetBodyBuffering( true );
			// Tell client not to cache generated response
			response.AddHeader( "Cache-Control: no-cache" );
			response.AddHeader( "Pragma: no-cache" );
			response.AddHeader( "Expires: now" );
			response.AddHeader( "Connection: close" );
			contentLength = body.BufferLength();
			
			// If GET, set body
			if( pb->request->GetMethod() == METHOD_GET ){
				if (
				((QueryData->has_error) && (QueryData->error_mode == 1)) ||
				(QueryData->has_error) && (!QueryData->has_dummy) && (QueryData->error_mode == 2)
				){
					response.AddHeader( kHEAD_TYPE, "text/html" );
					
					io_printf( &error_body, QueryData->error_report, QueryData->error_message);
					response.SetMessageBody( &error_body );
					error_body.Seek( 0, SEEK_SET );
					
				}else if(
				((QueryData->has_error) && (QueryData->has_dummy) && (QueryData->error_mode == 2)) ||
				((QueryData->has_error) && (QueryData->has_dummy) && (QueryData->error_mode == 3))
				){
					response.AddHeader( kHEAD_TYPE, QueryData->reports[0].content_type );
					
					QueryData->Format_BodyPart(&error_body, QueryData->reports[0].header);
					
					response.SetMessageBody( &error_body );
					error_body.Seek( 0, SEEK_SET );
					
					//(new BAlert("", QueryData->reports[0].content_type, "OK"))->Go();
				}else{
					response.AddHeader( kHEAD_TYPE, QueryData->reports[report_index].content_type);
					
					response.SetMessageBody( &body );
					body.Seek( 0, SEEK_SET );
				}
			}
			
			
			RHQdeltaTime = system_time() - RHQstartTime;
			int32 		RHQElapsed_mSeconds = RHQdeltaTime/1000;
			pb->Logprintf( "@# RHQuery complete HTTP response in %ld ms\n", 
				RHQElapsed_mSeconds );
				
			response.SetStatusLine( 200 );
			pb->Logprintf( "%ld Sending: %ld %s\n", pb->sn, contentLength, absPath.Path() );
			pb->Logprintf( "%ld Status-Line: %s\n", pb->sn, response.GetStatusLine() );
			int32	size;
			bigtime_t 		startTime, deltaTime;
			startTime = system_time();
			
			size = pb->request->SendReply( &response );
	
	
			pb->closeConnection = true;
			
			deltaTime = system_time() - startTime;
			int32 		bytesSeconds = (size*1000000)/deltaTime;
			pb->Logprintf( "%ld Sent: %ld bytes %ld ms %ld bytes/second\n", 
				pb->sn, size, deltaTime/1000, bytesSeconds );
			return B_OK;
		}else{
			pb->Logprintf("@# RHQ Going to stream\n");
			send_file(pb, QueryData->get_file_stream_path(report_index), QueryData->filetype);
		}
		
		delete QueryData;
	}else{
		response.SetHTMLMessage( 404 ); // Not Found
		pb->Logprintf( "%ld Status-Line: %s\n", pb->sn, response.GetStatusLine() );
		pb->request->SendReply( &response );
		return B_OK;
	}

}




status_t send_file( RequestPB *pb, char * filepath, char * filetype)
{
	
	HTTPResponse	response;
	int32			fieldSize = 1024;
	char			fieldValue[4096];
	char			headBuffer[1024];

	// Temporary buffers
	
	BFile			theFile;
	BPath			absPath;
	
	pb->closeConnection = true;


	absPath.SetTo(filepath);
	
	
	if( theFile.SetTo( absPath.Path(), B_READ_ONLY ) == B_NO_ERROR )
	{
		int32		statusCode = 200;
		time_t		modtime;
		theFile.GetModificationTime( &modtime );
		

		// Get file size
		int32		entityLength, contentLength;
		off_t size;
		
		theFile.GetSize( &size );
		entityLength = size;
		
		// Add Date header
		time_t			now;
		struct tm 		*brokentime;
		
		now = time( NULL );
		brTimeLock.Lock();
		brokentime = gmtime( &now );
		strftime(fieldValue, 256, kHTTP_DATE, brokentime);
		brTimeLock.Unlock();
		
		response.AddHeader( kHEAD_DATE, fieldValue );
		
		// Add last modified header
		brTimeLock.Lock();
		brokentime = gmtime (&modtime);
		strftime (fieldValue, fieldSize, kHTTP_DATE, brokentime);
		brTimeLock.Unlock();
		
		response.AddHeader( kHEAD_LAST_MODIFIED, fieldValue );
		
		// Add server name header
		response.AddHeader( kHEAD_SERVER, "RobinHood" );
		
		// Add Accept-Ranges header
		response.AddHeader( "Accept-Ranges: bytes" );
		
		// If "Range:" header, parse range
		ByteRangeSet	rangeSet;
		
		if( pb->request->FindHeader( kHEAD_RANGE, fieldValue, fieldSize ) )
		{
			rangeSet.AddByteRange( fieldValue );
			statusCode = 206;
		}
		
		if( statusCode == 206 ) // If partial content
		{
			contentLength = rangeSet.ContentLength( entityLength );
			response.SetContentLength( contentLength );
			if( rangeSet.CountRanges() == 1 ){
				response.AddHeader( kHEAD_CONTENT_RANGE, 
					rangeSet.ContentRangeString( headBuffer, 0, entityLength ) );
			}
		}
		else{ // Set content-length to entity-length
			contentLength = entityLength;
		}
	
		// Set content-length
		response.SetContentLength( contentLength );
		
		// Add content length header
		sprintf( headBuffer, "%s: %ld", kHEAD_LENGTH, contentLength );
		response.AddHeader( headBuffer );
		
		// Add mime type header
		response.AddHeader( kHEAD_TYPE, "audio/x-mpeg" );
		
		// Set message body
		BFile		*bodyFile;
		
		//if( pb->request->GetMethod() == METHOD_GET )
		//{
			response.SetMessageBody( &theFile );
			bodyFile = &theFile;
		//}
		//else // Don't send body on HEAD
		//	bodyFile = NULL;
		
		// Set status line
		response.SetStatusLine( statusCode );
		pb->Logprintf( "%ld Status-Line: %s\n", pb->sn, response.GetStatusLine() );
		pb->Logprintf( "%ld Sending: %ld %s\n", pb->sn, contentLength, absPath.Path() );
		
		bigtime_t 		startTime, deltaTime;
		startTime = system_time();

		statusCode = 200;
		response.SetStatusLine( statusCode );
		size = pb->request->SendReply( &response );


		deltaTime = system_time() - startTime;
		int32 		bytesSeconds = (size*1000000)/deltaTime;
		pb->Logprintf( "%ld Sent: %ld bytes %ld ms %ld bytes/second\n", 
			pb->sn, size, deltaTime/1000, bytesSeconds );

	}
	else
	{
		response.SetHTMLMessage( 404 ); // Not Found
		pb->Logprintf( "%ld Status-Line: %s\n", pb->sn, response.GetStatusLine() );
		pb->request->SendReply( &response );
	}
	return B_OK;
}
